/*
 * Copyright (c) 2015, Cloudera and Intel, Inc. All Rights Reserved.
 *
 * Cloudera, Inc. licenses this file to you under the Apache License,
 * Version 2.0 (the "License"). You may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * This software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for
 * the specific language governing permissions and limitations under the
 * License.
 */

package com.cloudera.oryx.app.speed.kmeans;

import javax.xml.bind.JAXBException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import com.typesafe.config.Config;
import org.apache.spark.api.java.JavaPairRDD;
import org.apache.spark.api.java.function.Function2;
import org.apache.spark.api.java.function.PairFunction;
import org.dmg.pmml.PMML;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.Tuple2;

import com.cloudera.oryx.app.common.fn.MLFunctions;
import com.cloudera.oryx.app.kmeans.ClusterInfo;
import com.cloudera.oryx.app.kmeans.KMeansPMMLUtils;
import com.cloudera.oryx.app.schema.InputSchema;
import com.cloudera.oryx.common.pmml.PMMLUtils;
import com.cloudera.oryx.common.text.TextUtils;
import com.cloudera.oryx.lambda.KeyMessage;
import com.cloudera.oryx.lambda.speed.SpeedModelManager;

public final class KMeansSpeedModelManager implements SpeedModelManager<String,String,String> {

  private static final Logger log = LoggerFactory.getLogger(KMeansSpeedModelManager.class);

  private KMeansSpeedModel model;
  private final InputSchema inputSchema;

  public KMeansSpeedModelManager(Config config) {
    inputSchema = new InputSchema(config);
  }

  @Override
  public void consume(Iterator<KeyMessage<String, String>> updateIterator) throws IOException {
    while (updateIterator.hasNext()) {
      KeyMessage<String, String> km = updateIterator.next();
      String key = km.getKey();
      String message = km.getMessage();
      switch (key) {
        case "UP":
          // do nothing, hearing our own updates
          break;
        case "MODEL":
          log.info("Loading new model");
          // New model
          PMML pmml;
          try {
            pmml = PMMLUtils.fromString(message);
          } catch (JAXBException e) {
            throw new IOException(e);
          }
          KMeansPMMLUtils.validatePMMLVsSchema(pmml, inputSchema);
          model = new KMeansSpeedModel(KMeansPMMLUtils.read(pmml));
          log.info("Model with {} clusters", model.getNumClusters());
          break;
        default:
          throw new IllegalStateException("Unexpected key " + key);
      }
    }

  }

  @Override
  public Iterable<String> buildUpdates(JavaPairRDD<String, String> newData) {
    if (model == null) {
      return Collections.emptyList();
    }

    List<Tuple2<Integer,Tuple2<double[],Long>>> updatedPoints =
        newData.values().map(MLFunctions.PARSE_FN)
            .mapToPair(new ToClusteredSums(inputSchema, model))
            .reduceByKey(new ReduceFeatureAndCountFn()).collect();

    List<String> updates = new ArrayList<>(updatedPoints.size());
    for (Tuple2<Integer,Tuple2<double[],Long>> pair : updatedPoints) {
      int clusterID = pair._1();
      double[] vectorSum = pair._2()._1();
      long count = pair._2()._2();

      for (int i = 0 ; i < vectorSum.length; i++) {
        vectorSum[i] /= count;
      }

      ClusterInfo clusterInfo = model.getCluster(clusterID);
      clusterInfo.update(vectorSum, count);
      model.setCluster(clusterID, clusterInfo);
      // add to updates
      updates.add(TextUtils.joinJSON(
          Arrays.asList(clusterID, clusterInfo.getCenter(), clusterInfo.getCount())));
    }

    return updates;
  }

  @Override
  public void close() {
    // do nothing
  }

  private static final class ToClusteredSums
      implements PairFunction<String[],Integer,Tuple2<double[],Long>> {

    private final InputSchema inputSchema;
    private final KMeansSpeedModel model;

    ToClusteredSums(InputSchema inputSchema, KMeansSpeedModel model) {
      this.inputSchema = inputSchema;
      // This assumes KMeansSpeedModel is smallish and trivial to serialize
      this.model = model;
    }

    @Override
    public Tuple2<Integer,Tuple2<double[],Long>> call(String[] data) {
      double[] featureVector = new double[inputSchema.getNumPredictors()];
      for (int featureIndex = 0; featureIndex < data.length; featureIndex++) {
        if (inputSchema.isActive(featureIndex)) {
          featureVector[inputSchema.featureToPredictorIndex(featureIndex)] =
              Double.parseDouble(data[featureIndex]);
        }
      }
      int closestClusterID = model.closestCluster(featureVector);
      return new Tuple2<>(closestClusterID, new Tuple2<>(featureVector, 1L));
    }
  }

  private static final class ReduceFeatureAndCountFn
      implements Function2<Tuple2<double[],Long>,Tuple2<double[],Long>,Tuple2<double[],Long>> {
    @Override
    public Tuple2<double[], Long> call(Tuple2<double[], Long> t1, Tuple2<double[], Long> t2) {
      double[] vec1 = t1._1();
      double[] vec2 = t2._1();
      // going to modify 1 in place
      for (int i = 0; i < vec1.length; i++) {
        vec1[i] += vec2[i];
      }
      long totalCount = t1._2() + t2._2();
      return new Tuple2<>(vec1, totalCount);
    }
  }

}
